"""
Define RPC functions needed to generate
"""

# [ <ret-type>, [<arg_types>] ]
func_table = [
    [ "string", ["string", "int"] ],
    [ "object", ["string"] ],
    [ "objlist", ["string", "int"] ],
    [ "json", ["string", "int"] ],
    [ "json", ["json"]],
]
